#include "core/File.h"

#include <stdio.h>

#include "ErrorSimulator.h"
#include "core/Logger.h"
#include "core/Utility.h"

static bool readOpenFile(FILE* file, FileContent* f, const char* path) {
    if(FAIL_STEP || fseek(file, 0, SEEK_END)) {
        REPORT(LOG_ERROR, "cannot seek file end of '%s'", path);
        return true;
    }
    long l = ftell(file);
    if(FAIL_STEP || l < 0) {
        REPORT(LOG_ERROR, "cannot tell file position of '%s'", path);
        return true;
    }
    f->length = (size_t)l;
    if(FAIL_STEP || fseek(file, 0, SEEK_SET)) {
        REPORT(LOG_ERROR, "cannot seek file start of '%s'", path);
        return true;
    }
    f->data = coreAllocate(f->length + 1);
    size_t read = fread(f->data, 1, f->length, file);
    f->data[f->length] = 0;
    if(FAIL_STEP || read != f->length) {
        REPORT(
            LOG_ERROR, "expected to read %zu bytes from '%s' but read %zu",
            f->length, path, read);
        return true;
    }
    return false;
}

bool readFile(FileContent* f, const char* path) {
    FILE* file = fopen(path, "rb");
    if(file == nullptr) {
        REPORT(LOG_ERROR, "cannot read file '%s'", path);
        return true;
    }
    bool r = readOpenFile(file, f, path);
    if(FAIL_STEP || fclose(file)) {
        REPORT(LOG_ERROR, "cannot close file '%s'", path);
        r = true;
    }
    if(r) {
        destroyFileContent(f);
    }
    return r;
}

void destroyFileContent(FileContent* f) {
    coreFree(f->data);
    *f = (FileContent){0};
}
