#include "core/Logger.h"

#include <stdarg.h>
#include <stdio.h>

LogLevel logLevel = LOG_DEBUG;
static ReportHandler reportHandler = nullptr;
static void* reportData = nullptr;

void setReportHandler(ReportHandler h, void* data) {
    reportHandler = h;
    reportData = data;
}

void callReportHandler(
    LogLevel l, const char* file, int line, const char* format, ...) {
    char buffer[512];

    va_list args;
    va_start(args, format);
    vsnprintf(buffer, sizeof(buffer), format, args);
    va_end(args);

    if(reportHandler != nullptr) {
        reportHandler(l, file, line, reportData, buffer);
    }
}

const char* getShortFileName(const char* s) {
    const char* r = s;
    while(*s != '\0') {
        if(*(s++) == '/') {
            r = s;
        }
    }
    return r;
}

void printLog(
    LogLevel l, const char* file, int line, const char* prefix, const char* tag,
    const char* format, ...) {
    if(logLevel < l) {
        return;
    }
    file = getShortFileName(file);
    fputs(prefix, stdout);
    fputs(tag, stdout);
    printf("%s:%d | ", file, line);

    va_list args;
    va_start(args, format);
    vprintf(format, args);
    va_end(args);

    puts(TERMINAL_RESET);
}
